from math import fabs

from graphene_django_extras import LimitOffsetGraphqlPagination
from graphene_django_extras.paginations.utils import _nonzero_int


class LimitOffsetOrderingGraphqlPagination(LimitOffsetGraphqlPagination):
    def paginate_queryset(self, qs, **kwargs):
        """
        The original method is not sorting when limit = None
        """
        order = kwargs.pop(self.ordering_param, None) or self.ordering

        if order:
            if "," in order:
                order = order.strip(",").replace(" ", "").split(",")
                if order.__len__() > 0:
                    qs = qs.order_by(*order)
            else:
                qs = qs.order_by(order)

        limit = _nonzero_int(
            kwargs.get(self.limit_query_param, None), strict=True, cutoff=self.max_limit
        )

        if limit is None:
            return qs

        offset = kwargs.get(self.offset_query_param, 0)

        return qs[offset: offset + fabs(limit)]



import graphene
from graphene import Argument, ID
from graphene_django.utils import maybe_queryset, is_valid_django_model
from graphene_django_extras import DjangoFilterListField, DjangoListObjectField, DjangoObjectField
from graphene_django_extras.base_types import DjangoListObjectBase
from graphene_django_extras.utils import queryset_factory, get_extra_filters


class OrderingMixin:
    """
    A mixin to add ordering feature to graphene_django_extra fields.

    Usage:

    query{
      books(ordering: "name"){
        id
        name
      }

      books(ordering: "name,-id"){
        id
        name
      }
    }
    """

    def __init__(self, *args, **kwargs):
        kwargs["args"] = {"ordering": Argument(graphene.String,
                                               description="String of comma-sepratated fields to order by.")}
        super().__init__(*args, **kwargs)

    @staticmethod
    def list_resolver(*args, **kwargs):
        qs = DjangoFilterListField.list_resolver(*args, **kwargs)
        if "ordering" in kwargs:
            ordering_fields = [field.strip() for field in kwargs["ordering"].split(",")]
            qs = qs.order_by(*ordering_fields)
        return qs


class DjangoFilterOrderingListField(OrderingMixin, DjangoFilterListField):
    """
    DjangoFilterListField with ordering feature.
    """


class PermissionsCheckMixin:
    def get_resolver(self, parent_resolver):
        if hasattr(parent_resolver, "permissions_check"):
            self.permission_check_method = parent_resolver
        return super().get_resolver(parent_resolver)

    def list_resolver(self, manager, filterset_class, filtering_args, root, info, **kwargs):
        output = super().list_resolver(manager, filterset_class, filtering_args, root, info, **kwargs)

        if hasattr(self, "permission_check_method"):
            self.permission_check_method(root, info, output.results, **kwargs)

        return output

    def object_resolver(self, manager, root, info, **kwargs):
        output = super().object_resolver(manager, root, info, **kwargs)

        if hasattr(self, "permission_check_method"):
            self.permission_check_method(output, info, **kwargs)

        return output


class DjangoNestableListObjectField(DjangoListObjectField):
    """
    Similar to DjangoListObjectField, except it can be nested into ManyToOneRel.
    Also, it can fetch queryset by property.
    """

    def __init__(self, _type, *args, fetch_fn=None, property_name=None, **kwargs):
        super().__init__(_type, *args, **kwargs)
        self.property_name = property_name
        self.fetch_fn = fetch_fn

    def list_resolver(self, manager, filterset_class, filtering_args, root, info, **kwargs):
        if self.fetch_fn is not None:
            qs = self.fetch_fn(info.context)
        elif self.property_name is not None and root and is_valid_django_model(root._meta.model):
            qs = getattr(root, self.property_name)
        else:
            qs = queryset_factory(manager, info.field_asts, info.fragments, **kwargs)

        filter_kwargs = {k: v for k, v in kwargs.items() if k in filtering_args}

        qs = filterset_class(data=filter_kwargs, queryset=qs, request=info.context).qs

        if root and is_valid_django_model(root._meta.model):
            extra_filters = get_extra_filters(root, manager.model)
            qs = qs.filter(**extra_filters)

        count = qs.count()
        results = maybe_queryset(qs)

        return DjangoListObjectBase(
            count=count,
            results=results,
            results_field_name=self.type._meta.results_field_name,
        )


class DjangoCustomObjectField(DjangoObjectField):
    def __init__(self, _type, *args, fetch_fn=None, id_arg=True, **kwargs):
        if id_arg:
            kwargs["id"] = ID(
                required=True, description="Django object unique identification field"
            )

        self.fetch_fn = fetch_fn
        super(DjangoObjectField, self).__init__(_type, *args, **kwargs)

    def object_resolver(self, manager, root, info, **kwargs):
        if self.fetch_fn is not None:
            pk = kwargs.get("id", None)
            if pk is not None:
                pk = int(pk)
            return self.fetch_fn(info.context, pk)
        return super().object_resolver(manager, root, info, **kwargs)


class DjangoObjectPermissionsField(PermissionsCheckMixin, DjangoCustomObjectField):
    pass


class DjangoNestableListObjectPermissionsField(PermissionsCheckMixin, DjangoNestableListObjectField):
    pass



import django_describer.adapters.graphql.main
import django_describer.adapters.graphql.modifying
import django_describer.adapters.graphql.fields
import django_describer.adapters.graphql.retrieving
import django_describer.adapters.graphql.pagination
import django_describer.adapters.graphql.converter



from enum import Enum

from django_describer.permissions import AllowAll
from .utils import ensure_tuple, set_param_if_unset, get_object_or_raise, build_extra_fields, determine_fields


class ActionName(Enum):
    CREATE = "create"
    UPDATE = "update"
    DELETE = "delete"
    LIST = "list"
    DETAIL = "detail"

    @classmethod
    def values(cls):
        return set(item.value for item in cls)


class BaseAction:
    read_only = False
    has_model = True

    def __init__(self, permissions=None):
        self.permissions = ensure_tuple(permissions)
        self._describer = None
        self._name = None

    def set_describer(self, describer):
        set_param_if_unset(self, "_describer", describer)

    def set_name(self, name):
        set_param_if_unset(self, "_name", name)

    def get_name(self, capitalized=False, camelcase=False):
        if self._name is None:
            raise ValueError("_name is not set.")

        model = ""

        if self._describer is not None:
            model = self._describer.model.__name__

        if camelcase:
            name = model + self._name.capitalize()
        else:
            name = "{}_{}".format(model, self._name)

        if capitalized:
            name = name.capitalize()

        return name

    def get_permissions(self):
        if self.permissions:
            return self.permissions
        if self._describer is not None:
            self._describer.get_default_action_permissions()
        return AllowAll,

    def convert(self, to, **kwargs):
        raise NotImplementedError


class RetrieveAction(BaseAction):
    read_only = True

    def __init__(self, permissions=None, fetch_fn=None):
        super().__init__(permissions=permissions)
        self.fetch_fn = fetch_fn

    def get_fetch_fn(self):
        return self.fetch_fn or self.get_default_fetch_fn()

    def get_default_fetch_fn(self):
        raise NotImplementedError


class ListAction(RetrieveAction):
    def get_default_fetch_fn(self):
        def fn(request):
            return self._describer.model.objects.all()
        return fn

    def convert(self, to, **kwargs):
        return to.list_action(self, **kwargs)


class DetailAction(RetrieveAction):
    def __init__(self, permissions=None, fetch_fn=None, id_arg=True):
        super().__init__(permissions=permissions, fetch_fn=fetch_fn)
        self.id_arg = id_arg

    def get_default_fetch_fn(self):
        def fn(request, pk):
            return get_object_or_raise(self._describer.model, pk)
        return fn

    def convert(self, to, **kwargs):
        return to.detail_action(self, **kwargs)


class ModifyAction(BaseAction):
    def __init__(self, permissions=None, only_fields=None, exclude_fields=None, extra_fields=None, exec_fn=None,
                 return_fields=None):
        super().__init__(permissions=permissions)
        self.only_fields = ensure_tuple(only_fields, convert_none=False)
        self.exclude_fields = ensure_tuple(exclude_fields, convert_none=False)
        self.extra_fields = build_extra_fields(extra_fields)
        self.exec_fn = exec_fn
        self.return_fields = build_extra_fields(return_fields)

    def get_exec_fn(self):
        return self.exec_fn or self.get_default_exec_fn()

    def get_default_exec_fn(self):
        raise NotImplementedError

    def get_return_fields(self):
        return self.return_fields or build_extra_fields(self.get_default_return_fields())

    def get_default_return_fields(self):
        raise NotImplementedError

    def determine_fields(self):
        return determine_fields(self._describer.model, self.only_fields, self.exclude_fields, no_reverse=True)


class CreateAction(ModifyAction):
    def get_default_exec_fn(self):
        def fn(request, data):
            obj = self._describer.model(**data)
            obj.save()
            return {"object": obj}
        return fn

    def get_default_return_fields(self):
        return {"object": self._describer.model}

    def determine_fields(self):
        fields = super().determine_fields()
        fields_without_id = tuple(f for f in fields if f != "id")
        return fields_without_id

    def convert(self, to, **kwargs):
        return to.create_action(self, **kwargs)


class FetchModifyAction(ModifyAction):
    def __init__(self, permissions=None, only_fields=None, exclude_fields=None, extra_fields=None, exec_fn=None,
                 return_fields=None, fetch_fn=None):
        super().__init__(permissions=permissions, only_fields=only_fields, exclude_fields=exclude_fields,
                         extra_fields=extra_fields, exec_fn=exec_fn, return_fields=return_fields)
        self.fetch_fn = fetch_fn

    def get_fetch_fn(self):
        return self.fetch_fn or self.get_default_fetch_fn()

    def get_default_fetch_fn(self):
        def fn(request, pk):
            return get_object_or_raise(self._describer.model, pk)
        return fn


class UpdateAction(FetchModifyAction):
    def get_default_exec_fn(self):
        def fn(request, instance, data):
            for k, v in data:
                setattr(instance, k, v)
            instance.save()
            return instance
        return fn

    def get_default_return_fields(self):
        return {"object": self._describer.model}

    def determine_fields(self):
        fields = super().determine_fields()
        if "id" not in fields:
            fields = fields + ("id",)
        return fields

    def convert(self, to, **kwargs):
        return to.update_action(self, **kwargs)


class DeleteAction(FetchModifyAction):
    def __init__(self, permissions=None, extra_fields=None, exec_fn=None, return_fields=None, fetch_fn=None):
        super().__init__(permissions=permissions, extra_fields=extra_fields, exec_fn=exec_fn,
                         return_fields=return_fields, fetch_fn=fetch_fn)

    def get_default_exec_fn(self):
        def fn(request, instance, data):
            instance.delete()
            return instance
        return fn

    def get_default_return_fields(self):
        return {"object": self._describer.model}

    def determine_fields(self):
        return "id",

    def convert(self, to, **kwargs):
        return to.delete_action(self, **kwargs)


class CustomAction(ModifyAction):
    has_model = False

    def __init__(self, input_type, return_fields, exec_fn, permissions=None):
        super().__init__(permissions=permissions, exec_fn=exec_fn, return_fields=return_fields)
        self.input_type = input_type

    def get_default_exec_fn(self):
        raise ValueError("No default exec_fn, you need to provide one.")

    def get_default_return_fields(self):
        raise ValueError("No default return fields, you need to provide some.")

    def convert(self, to, **kwargs):
        return to.custom_action(self, **kwargs)



import django.db.models
import graphene
from graphene_django_extras import DjangoObjectType, DjangoListObjectType
from graphene_django_extras.settings import graphql_api_settings

from django_describer.adapters.utils import register_action_name
from ...datatypes import String, Integer, Float, Boolean, NullType, get_instantiated_type
from .converter import convert_local_fields
from .pagination import LimitOffsetOrderingGraphqlPagination
from ...utils import field_names, get_all_model_fields


# mapping of alien types to django_describer ones
_reverse_field_map = {
    # Django types
    django.db.models.fields.CharField: String,
    django.db.models.fields.TextField: String,
    django.db.models.fields.IntegerField: Integer,
    django.db.models.fields.FloatField: Float,
    django.db.models.fields.BooleanField: Boolean,
    django.db.models.fields.AutoField: Integer,
}


class Query:
    """
    A type for local graphene Query classes. Intentionally blank, serves only for type checking.
    """


def create_type_class(describer):
    """
    Creates a DjangoObjectType and a DjangoListObjectType for a model, and links them together.
    Must be in this order!!!
    """
    type_meta = type(
        "Meta",
        (object,),
        {
            "model": describer.model,
            "filter_fields": create_filter_fields(describer),
            "only_fields": describer.get_fields(),
        }
    )

    type_class = type(
        "{}Type".format(describer.model.__name__),
        (DjangoObjectType,),
        {
            "Meta": type_meta,
            "get_list_type": lambda: type_list_class,
            **convert_local_fields(describer.model, describer.get_fields())
        }
    )

    type_list_meta = type(
        "Meta",
        (object,),
        {
            "model": describer.model,
            "pagination": LimitOffsetOrderingGraphqlPagination(
                default_limit=describer.default_page_size or graphql_api_settings.DEFAULT_PAGE_SIZE,
                max_limit=describer.max_page_size or graphql_api_settings.DEFAULT_PAGE_SIZE
            ),
        }
    )

    type_list_class = type(
        "{}ListType".format(describer.model.__name__),
        (DjangoListObjectType,),
        {
            "Meta": type_list_meta,
        }
    )

    return type_class


def create_filter_fields(describer):
    """
    Creates dictionary of filters based on field types.
    """
    filter_fields = {}
    field_names = (f.name for f in describer.model._meta.fields)

    for field_name in field_names:

        # get the filters for each field based on their types
        field_type = _reverse_field_map.get(describer.model._meta.get_field(field_name).__class__, NullType)

        # add the filters to the output
        filter_fields[field_name] = field_type.filters()

    return filter_fields


def create_query_class(adapter, actions):
    """
    Creates a Query class, featuring listing and detail methods.
    """

    attrs = {}

    for action in actions:
        if not action.read_only:
            continue

        name = action.get_name()
        register_action_name(adapter, name)

        attrs[name] = action.convert(adapter)

        if action.get_permissions():
            attrs["resolve_{}".format(name)] = create_permissions_check_method(
                permission_classes=action.get_permissions())

    query_class = type(
        "Query",
        (Query,),
        attrs,
    )

    return query_class


def add_permissions_to_type_class(describer, type_class):
    """
    Adds permissions to the given DjangoObjectType class.
    """
    for field in describer.get_fields():
        permissions = None

        if field in describer.get_field_permissions():
            permissions = describer.get_field_permissions()[field]
        elif describer.get_default_field_permissions():
            permissions = describer.get_default_field_permissions()

        if permissions:
            setattr(type_class,
                    "resolve_{}".format(field),
                    create_permissions_check_method(field, permissions))


def add_extra_fields_to_type_class(adapter, describer, type_class):
    """
    Adds extra fields to the given DjangoObjectType class. Ensures no base fields get overwritten.
    """
    existing_fields = field_names(get_all_model_fields(describer.model))

    for field_name, return_type in describer.get_extra_fields().items():
        if field_name in existing_fields:
            raise ValueError("This field already exists.")

        type_class._meta.fields[field_name] = get_instantiated_type(return_type).convert(
            adapter, property_name=field_name)


def create_global_query_class(query_classes, non_model_query_class):
    return type("Query", query_classes.values() + (non_model_query_class,) + (graphene.ObjectType,), {})


def create_permissions_check_method(field_name=None, permission_classes=()):
    """
    Generator of methods to check permissions for both ListFields and Fields.
    """
    def method(root, info, results=None, **kwargs):
        for permission_class in permission_classes:
            pc = permission_class(info.context, obj=root, qs=results)
            if not pc.has_permission():
                raise PermissionError(pc.error_message())

        # return only for non-list Fields
        if field_name and hasattr(root, field_name):
            return getattr(root, field_name)

    # necessary flag for ListFields
    method.permissions_check = True
    return method



import graphene
from graphene import ObjectType
from graphene_django_extras import DjangoInputObjectType

from django_describer.adapters.utils import register_action_name
from django_describer.datatypes import get_instantiated_type


def create_mutation_classes(adapter, actions):
    mutation_classes = []
    for action in actions:
        if action.read_only:
            continue
        mutation_classes.append(create_mutation_class(adapter, action, has_model=action.has_model))
    return mutation_classes


def create_mutate_method(action, has_model=True):
    """
    Creates the mutate method based on fn. Adds permissions as well.
    """

    @classmethod
    def mutate(cls, root, info, *args, **kwargs):
        obj = None
        if has_model and "id" in kwargs["data"]:
            obj = action.get_fetch_fn()(info.context, kwargs["data"]["id"])

        for permission_class in action.get_permissions():
            pc = permission_class(info.context, obj=obj, data=kwargs["data"])
            if not pc.has_permission():
                raise PermissionError(pc.error_message())

        if obj is not None:
            return action.get_exec_fn()(info.context, obj, kwargs["data"])
        return action.get_exec_fn()(info.context, kwargs["data"])

    return mutate


def create_return_fields(adapter, action):
    ret = {}
    for name, type in action.get_return_fields().items():
        ret[name] = get_instantiated_type(type).convert(adapter)
    return ret


def create_mutation_class(adapter, action, has_model=True):
    if has_model:
        input_meta = type(
            "Meta",
            (object,),
            {
                "model": action._describer.model,
                "only_fields": action.determine_fields(),
                "input_for": action._name,
            }
        )

        input_class = type(
            "{}{}Input".format(action._describer.model.__name__, action._name.capitalize()),
            (DjangoInputObjectType,),
            {
                "Meta": input_meta
            }
        )

        # add extra fields to input type
        for name, return_type in action.extra_fields.items():
            if name in input_class._meta.input_fields:
                raise ValueError("Duplicate field: `{}`".format(name))
            input_class._meta.input_fields[name] = get_instantiated_type(return_type).convert(adapter, input=True)
        input_type = input_class(required=True)
    else:
        input_type = get_instantiated_type(action.input_type).convert(adapter, input=True)

    arguments_class = type(
        "Arguments",
        (object,),
        {
            "data": input_type
        }
    )

    return_fields = create_return_fields(adapter, action)

    mutation_class = type(
        "{}Mutation".format(action.get_name(camelcase=True)),
        (graphene.Mutation,),
        {
            **return_fields,
            "Arguments": arguments_class,
            "mutate": create_mutate_method(action, has_model=has_model),
        }
    )

    name = action.get_name()
    register_action_name(adapter, name)
    mutation_class._name = name

    return mutation_class


def create_global_mutation_class(mutation_classes, non_model_mutation_classes):
    attrs = {}
    for model, mutations in mutation_classes.items():
        for mutation in mutations:
            attrs[mutation._name] = mutation.Field()

    for mutation in non_model_mutation_classes:
        attrs[mutation._name] = mutation.Field()

    mutation_class = type(
        "Mutation",
        (ObjectType,),
        attrs
    )

    return mutation_class



from django.db.models import ManyToOneRel, ManyToManyRel

from .datatypes import get_instantiated_type


class AttrDict(dict):
    """
    A dictionary with keys accessible as attributes
    """
    def __getattr__(self, item):
        return self[item]

    def keys(self):
        return tuple(super().keys())

    def values(self):
        return tuple(super().values())


def __reverse_fields(model, local_field_names):
    for name, attr in model.__dict__.items():
        # Don't duplicate any local fields
        if name in local_field_names:
            continue

        # Django =>1.9 uses 'rel', django <1.9 uses 'related'
        related = getattr(attr, "rel", None) or getattr(attr, "related", None)
        if isinstance(related, ManyToOneRel):
            yield (name, related)
        elif isinstance(related, ManyToManyRel) and not related.symmetrical:
            yield (name, related)


def _reverse_fields(model, local_field_names):
    return list(__reverse_fields(model, local_field_names))


def get_local_fields(model):
    return [
        (field.name, field)
        for field in sorted(
            list(model._meta.fields) + list(model._meta.local_many_to_many)
        )
    ]


def get_reverse_fields(model):
    local_fields = get_local_fields(model)

    # Make sure we don't duplicate local fields with "reverse" version
    local_field_names = [field[0] for field in local_fields]
    return _reverse_fields(model, local_field_names)


def get_all_model_fields(model):
    return get_local_fields(model) + get_reverse_fields(model)


def field_names(field_tuple):
    return tuple(f[0] for f in field_tuple)


def determine_fields(model, only_fields, exclude_fields, no_reverse=False):
    if no_reverse:
        all_fields = field_names(get_local_fields(model))
    else:
        all_fields = field_names(get_all_model_fields(model))

    if only_fields is None and exclude_fields is None:
        exclude_fields = ()

    if only_fields is not None and exclude_fields is not None:
        raise ValueError("Cannot define both only_fields and exclude_fields.")

    if only_fields is not None:
        if not isinstance(only_fields, (list, tuple)):
            only_fields = only_fields,
        for field in only_fields:
            if field not in all_fields:
                raise ValueError("Unknown field: {}.".format(field))
        return only_fields

    for field in exclude_fields:
        if not isinstance(exclude_fields, (list, tuple)):
            only_fields = only_fields,
        if field not in all_fields:
            raise ValueError("Unknown field: {}.".format(field))
    return tuple(f for f in all_fields if f not in exclude_fields)


def model_plural_name(model):
    return str(model._meta.verbose_name_plural)


def model_singular_name(model):
    return str(model._meta.verbose_name)


def get_object_or_none(model, pk):
    qs = model.objects.filter(pk=pk)
    if not qs.exists():
        return None
    return qs.get()


def get_object_or_raise(model, pk):
    qs = model.objects.filter(pk=pk)
    if not qs.exists():
        raise ValueError("`{}` with pk={} does not exist.".format(model, pk))
    return qs.get()


def ensure_tuple(maybe_tuple, convert_none=True):
    if maybe_tuple is None:
        if convert_none:
            return tuple()
        return None
    if isinstance(maybe_tuple, (list, tuple)):
        return tuple(maybe_tuple)
    return maybe_tuple,


def build_extra_fields(extra_fields):
    if extra_fields is None:
        return {}
    ret = {}
    for name, return_type in extra_fields.items():
        ret[name] = get_instantiated_type(return_type)
    return ret


def build_field_permissions(field_permissions):
    if field_permissions is None:
        return {}
    ret = {}
    for field, permissions in field_permissions.items():
        ret[field] = ensure_tuple(permissions)
    return ret


def set_param_if_unset(obj, param, value):
    if hasattr(obj, param) and getattr(obj, param) is not None:
        raise ValueError("`{}` is already set.".format(param))
    setattr(obj, param, value)



from collections import OrderedDict
from copy import deepcopy

import graphene
from django.contrib.contenttypes.fields import GenericRel
import django.db.models
from graphene_django.fields import DjangoListField
from graphene_django_extras.converter import convert_django_field, convert_django_field_with_choices
from graphene_django_extras.registry import get_global_registry
from graphene_django_extras.utils import is_required

from .fields import DjangoNestableListObjectPermissionsField
from ...utils import get_local_fields


@convert_django_field.register(GenericRel)
@convert_django_field.register(django.db.models.ManyToManyRel)
@convert_django_field.register(django.db.models.ManyToOneRel)
def convert_many_rel_to_djangomodel(
    field, registry=None, input_flag=None, nested_field=False
):
    """
    An override of the original convert function. Takes into account improved fields with ordering and pagination.
    """

    model = field.related_model

    def dynamic_type():
        if input_flag and not nested_field:
            return DjangoListField(graphene.ID)
        else:
            _type = registry.get_type_for_model(model, for_input=input_flag)

            # get list type of the object if it has one
            if hasattr(_type, "get_list_type"):
                _type = _type.get_list_type()

            if not _type:
                return
            elif input_flag and nested_field:
                return DjangoListField(_type)
            elif _type._meta.filter_fields or _type._meta.filterset_class:
                # return nested relations as a field with pagination
                return DjangoNestableListObjectPermissionsField(
                    _type,
                    required=is_required(field) and input_flag == "create",
                    filterset_class=_type._meta.filterset_class,
                )
            else:
                return DjangoListField(
                    _type, required=is_required(field) and input_flag == "create"
                )

    return graphene.Dynamic(dynamic_type)


def convert_local_fields(model, convertable_fields):
    """
    Converts user-specified model fields as if they were nullable.
    Also transforms ID to integer field, since it is (usually) an integer.
    """

    # define how each field shall be converted
    field_conversions = {
        django.db.models.fields.AutoField: django.db.models.IntegerField
    }

    fields = [[name, deepcopy(field)] for name, field in get_local_fields(model)]

    for i in range(len(fields)):
        # change a field's type if necessary
        new_field = field_conversions.get(fields[i][1].__class__, None)
        if new_field is not None:
            fields[i][1] = new_field()

        # set all fields as nullable
        fields[i][1].null = True

    # convert the fields
    ret = OrderedDict()

    for name, field in fields:
        if name not in convertable_fields:
            continue
        converted = convert_django_field_with_choices(field, get_global_registry())
        ret[name] = converted

    return ret


def register_action_name(adapter, name):
    if not hasattr(adapter, "_action_names"):
        adapter._action_names = set()
    if name in adapter._action_names:
        raise ValueError("Duplicate action name: `{}`".format(name))
    adapter._action_names.add(name)


non_model_actions = []


def register_non_model_action(name, action):
    action.set_name(name)
    non_model_actions.append(action)


def generate(adapter):
    return adapter().generate()



from django.utils.translation import ugettext_lazy as _

from .utils import AttrDict


class BasePermission:
    def permission_statement(self):
        raise NotImplementedError

    def has_permission(self):
        for clas in reversed(self.__class__.__mro__):
            if clas in (object, BasePermission, Permission, OrResolver):
                continue
            if not clas.permission_statement(self):
                return False
        return True

    def error_message(self):
        return _("You don't have permission to do this.")


class Permission(BasePermission):
    def __init__(self, request, obj=None, data=None, qs=None):
        self.request = request
        self.obj = obj
        self.data = AttrDict(data) if data else AttrDict()
        self.qs = qs


class OrResolver(Permission):
    def __init__(self, permission_classes, request, obj=None, data=None, qs=None):
        super().__init__(request=request, obj=obj, data=data, qs=qs)
        self.permission_classes = permission_classes
        self.errors = []

    def has_permission(self):
        for permission_class in self.permission_classes:
            pc = permission_class(self.request, obj=self.obj, data=self.data, qs=self.qs)
            if pc.has_permission():
                return True
            self.errors.append(pc.error_message())
        return False

    def error_message(self):
        err = str(self.errors[0])
        for i in range(1, len(self.errors)):
            err += " "
            err += str(_("OR"))
            err += " "
            err += str(self.errors[i])
        return err


class Or:
    def __init__(self, *permissions):
        self.permissions = permissions

    def __call__(self, request, obj=None, data=None, qs=None):
        return OrResolver(self.permissions, request, obj=obj, data=data, qs=qs)


class AllowAll(Permission):
    def permission_statement(self):
        return True


class AllowNone(Permission):
    def permission_statement(self):
        return False


class IsAuthenticated(Permission):
    def permission_statement(self):
        return self.request.user and self.request.user.is_authenticated

    def error_message(self):
        return _("Log in to access this.")



class Adapter:
    def string_type(self, type, **kwargs):
        raise NotImplementedError

    def integer_type(self, type, **kwargs):
        raise NotImplementedError

    def float_type(self, type, **kwargs):
        raise NotImplementedError

    def id_type(self, type, **kwargs):
        raise NotImplementedError

    def boolean_type(self, type, **kwargs):
        raise NotImplementedError

    def queryset_type(self, type, **kwargs):
        raise NotImplementedError

    def model_type(self, type, **kwargs):
        raise NotImplementedError

    def composite_type(self, type, **kwargs):
        raise NotImplementedError

    def list_action(self, action, **kwargs):
        raise NotImplementedError

    def detail_action(self, action, **kwargs):
        raise NotImplementedError

    def generate(self, **kwargs):
        raise NotImplementedError



import django_describer.actions
import django_describer.datatypes
import django_describer.permissions
import django_describer.describers
import django_describer.utils
import django_describer.adapters

name = "django_describer"



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="django_describer",
    version="0.0.13",
    author="Karel Jilek",
    author_email="los.karlosss@gmail.com",
    description="A tool for automated generation of several APIs from a Django webapp.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/karlosss/django_describer",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "django",
        "django-filter",
        "graphene",
        "graphene-django",
        "graphene-django-extras",
    ],
    python_requires='>=3.6',
)



from inspect import isclass

from django.db.models import Model


class Type:
    def __init__(self, required=True, **kwargs):
        self.kwargs = kwargs
        self.kwargs["required"] = required

    def convert(self, to, **kwargs):
        raise NotImplementedError

    @staticmethod
    def filters():
        return ()


class String(Type):
    def convert(self, to, **kwargs):
        return to.string_type(self, **kwargs)

    @staticmethod
    def filters():
        return (
            "contains",
            "endswith",
            "exact",
            "icontains",
            "iendswith",
            "iexact",
            "iregex",
            "isnull",
            "istartswith",
            "regex",
            "startswith",
        )


class Integer(Type):
    def convert(self, to, **kwargs):
        return to.integer_type(self, **kwargs)

    @staticmethod
    def filters():
        return (
            "exact",
            "gt",
            "gte",
            "isnull",
            "lt",
            "lte",
        )


class Float(Type):
    def convert(self, to, **kwargs):
        return to.float_type(self, **kwargs)


class ID(Type):
    def convert(self, to, **kwargs):
        return to.id_type(self, **kwargs)


class Boolean(Type):
    def convert(self, to, **kwargs):
        return to.boolean_type(self, **kwargs)


class ModelType(Type):
    def __init__(self, model, **kwargs):
        self.model = model
        super().__init__(**kwargs)

    def convert(self, to, **kwargs):
        return to.model_type(self, **kwargs)


model_type_mapping = {}  # key: Model, value: ModelType


class QuerySet(Type):
    def __init__(self, of_type, **kwargs):
        self.of_type = of_type
        super().__init__(**kwargs)

    @property
    def type(self):
        return model_type_mapping[self.of_type] or self.of_type

    def convert(self, to, **kwargs):
        return to.queryset_type(self, **kwargs)


class NullType(Type):
    """
    Null object for types. convert() raises an exception. Useful when converting an unknown type for filters.
    """

    def convert(self, to, **kwargs):
        raise ValueError("Trying to convert NullType.")


class CompositeType(Type):
    """
    Technically, this is an Object type without Django model.
    """

    def __init__(self, type_name, field_map, **kwargs):
        super().__init__(**kwargs)
        self.type_name = type_name
        self.field_map = field_map  # key: field name, value: field type

    def convert(self, to, **kwargs):
        return to.composite_type(self, **kwargs)


def get_instantiated_type(maybe_type):
    if maybe_type in model_type_mapping:
        return model_type_mapping[maybe_type]
    if isclass(maybe_type):
        if issubclass(maybe_type, Model):
            raise ValueError(
                "`{m}` has not been registered yet. Please use `ModelType({m})` instead.".format(m=maybe_type.__name__))
        else:
            return maybe_type()
    return maybe_type



import logging

import graphene
from graphene.types.utils import get_field_as
from graphene_django.views import GraphQLView

from django_describer.adapters.utils import non_model_actions
from ..base import Adapter
from .fields import DjangoNestableListObjectPermissionsField, DjangoObjectPermissionsField
from ...datatypes import get_instantiated_type
from ...describers import get_describers
from ...utils import AttrDict
from .retrieving import create_type_class, add_extra_fields_to_type_class, add_permissions_to_type_class, \
    create_query_class, create_global_query_class
from .modifying import create_mutation_classes, create_global_mutation_class

create_class = type


class GraphQL(Adapter):
    def _convert_primitive_type(self, type, **kwargs):
        """
        A helper for converting primitive types to Graphene.
        """
        if "input" in kwargs and kwargs["input"]:
            return type
        return get_field_as(type, _as=graphene.Field)

    def string_type(self, type, **kwargs):
        return self._convert_primitive_type(graphene.types.String(required=type.kwargs["required"]), **kwargs)

    def integer_type(self, type, **kwargs):
        return self._convert_primitive_type(graphene.types.Int(required=type.kwargs["required"]), **kwargs)

    def float_type(self, type, **kwargs):
        return self._convert_primitive_type(graphene.types.Float(required=type.kwargs["required"]), **kwargs)

    def id_type(self, type, **kwargs):
        return self._convert_primitive_type(graphene.types.ID(required=type.kwargs["required"]), **kwargs)

    def boolean_type(self, type, **kwargs):
        return self._convert_primitive_type(graphene.types.Boolean(required=type.kwargs["required"]), **kwargs)

    def queryset_type(self, type, **kwargs):
        """
        Returns a field for listing. If the data shall be fetched from a queryset returned by its property,
        the property name is passed via kwargs.
        """
        if "input" in kwargs and kwargs["input"]:
            raise ValueError("Cannot convert QuerySet as input parameter.")

        property_name = kwargs.get("property_name", None)

        return graphene.Dynamic(lambda: DjangoNestableListObjectPermissionsField(
            type.type.convert(self, list=True),  # listing type is derived from the type passed as the of_type argument
            property_name=property_name))

    def model_type(self, type, **kwargs):
        """
        Returns a DjangoListObjectType if converting for lists, Field otherwise.
        """
        if "list" in kwargs and kwargs["list"]:
            return self.type_classes[type.model].get_list_type()

        if "input" in kwargs and kwargs["input"]:
            raise ValueError("Cannot convert ModelType as input parameter.")

        return graphene.Dynamic(lambda: graphene.Field(self.type_classes[type.model]))

    def composite_type(self, type, **kwargs):
        attrs = {}
        for name, return_type in type.field_map.items():
            attrs[name] = get_instantiated_type(return_type).convert(self, input="input" in kwargs and kwargs["input"])

        if "input" in kwargs and kwargs["input"]:
            return create_class(
                "{}Input".format(type.type_name),
                (graphene.InputObjectType,),
                attrs,
            )(required=type.kwargs["required"])

        type_class = create_class(
            "{}ObjectType".format(type.type_name),
            (graphene.ObjectType,),
            attrs
        )
        return graphene.Field(type_class)

    def list_action(self, action, **kwargs):
        return DjangoNestableListObjectPermissionsField(
            self.type_classes[action._describer.model].get_list_type(), fetch_fn=action.get_fetch_fn())

    def detail_action(self, action, **kwargs):
        return DjangoObjectPermissionsField(self.type_classes[action._describer.model], fetch_fn=action.get_fetch_fn(),
                                            id_arg=action.id_arg)

    def generate(self):
        # silence GraphQL exception logger
        logging.getLogger("graphql.execution.utils").setLevel(logging.CRITICAL)

        describers = get_describers()

        self.type_classes = AttrDict()  # key: Model, value: DjangoObjectType
        self.query_classes = AttrDict()  # key: Model, value: Query
        self.mutation_classes = AttrDict()  # key: Model, value: Mutation

        for describer in describers:
            # create a DjangoObjectType for the model
            self.type_classes[describer.model] = create_type_class(describer)

            # add extra fields to each DjangoObjectType class
            add_extra_fields_to_type_class(self, describer, self.type_classes[describer.model])

            # add permissions to each DjangoObjectType class (object fields)
            add_permissions_to_type_class(describer, self.type_classes[describer.model])

        for describer in describers:
            # create a Query class for each model (need to create all of them first)
            self.query_classes[describer.model] = create_query_class(self, describer.get_actions())

            # create mutation classes for the describer, including permissions and extra fields
            self.mutation_classes[describer.model] = create_mutation_classes(self, describer.get_actions())

        non_model_query_class = create_query_class(self, non_model_actions)
        non_model_mutation_classes = create_mutation_classes(self, non_model_actions)

        # create GraphQL schema
        schema = graphene.Schema(
            query=create_global_query_class(self.query_classes, non_model_query_class),
            mutation=create_global_mutation_class(self.mutation_classes, non_model_mutation_classes)
        )

        # create GraphQL view
        return GraphQLView.as_view(graphiql=True, schema=schema)



from copy import deepcopy

from .datatypes import model_type_mapping, ModelType
from .utils import determine_fields, ensure_tuple, build_field_permissions, build_extra_fields
from .actions import ListAction, DetailAction, ActionName, CreateAction, UpdateAction, DeleteAction


def get_describers():
    return tuple(DescriberMeta.all_describers.values())


class DescriberMeta(type):
    all_describers = {}
    """
    Automatically registers each describer and prevents multiple describers for a model.
    """
    def __new__(mcs, name, bases, attrs, **kwargs):
        cls = super().__new__(mcs, name, bases, attrs)
        if not cls.is_abstract():

            # create a ModelType off of the model
            if cls.model in model_type_mapping:
                raise ValueError("Multiple describers for `{}` model".format(cls.model.__name__))
            model_type_mapping[cls.model] = ModelType(cls.model)

            # save the describer
            DescriberMeta.all_describers[cls.model] = cls

            cls._fields = determine_fields(cls.model,
                                           ensure_tuple(cls.only_fields, convert_none=False),
                                           ensure_tuple(cls.exclude_fields, convert_none=False))
            cls._field_permissions = build_field_permissions(cls.field_permissions)
            cls._default_field_permissions = ensure_tuple(cls.default_field_permissions)
            cls._default_action_permissions = ensure_tuple(cls.default_action_permissions)

            cls._actions = []

            if cls.list_action is not None:
                cls.list_action = deepcopy(cls.list_action)
                cls.list_action.set_describer(cls)
                cls.list_action.set_name(ActionName.LIST.name)
                cls._actions.append(cls.list_action)

            if cls.detail_action is not None:
                cls.detail_action = deepcopy(cls.detail_action)
                cls.detail_action.set_describer(cls)
                cls.detail_action.set_name(ActionName.DETAIL.name)
                cls._actions.append(cls.detail_action)

            if cls.create_action is not None:
                cls.create_action = deepcopy(cls.create_action)
                cls.create_action.set_describer(cls)
                cls.create_action.set_name(ActionName.CREATE.name)
                cls._actions.append(cls.create_action)

            if cls.update_action is not None:
                cls.update_action = deepcopy(cls.update_action)
                cls.update_action.set_describer(cls)
                cls.update_action.set_name(ActionName.UPDATE.name)
                cls._actions.append(cls.update_action)

            if cls.delete_action is not None:
                cls.delete_action = deepcopy(cls.delete_action)
                cls.delete_action.set_describer(cls)
                cls.delete_action.set_name(ActionName.DELETE.name)
                cls._actions.append(cls.delete_action)

            for name, action in cls.extra_actions.items():
                if name in ActionName.values():
                    raise ValueError("`{}` is a reserved action name.".format(name))
                action.set_describer(cls)
                action.set_name(name)
                cls._actions.append(action)

        return cls


class Describer(metaclass=DescriberMeta):
    @classmethod
    def is_abstract(cls):
        return cls.model is None

    @classmethod
    def get_fields(cls):
        return cls._fields

    @classmethod
    def get_field_permissions(cls):
        return cls._field_permissions

    @classmethod
    def get_default_field_permissions(cls):
        return cls._default_field_permissions

    @classmethod
    def get_default_action_permissions(cls):
        return cls._default_action_permissions

    @classmethod
    def get_actions(cls):
        return cls._actions

    @classmethod
    def get_extra_fields(cls):
        if not hasattr(cls, "_extra_fields"):
            cls._extra_fields = build_extra_fields(cls.extra_fields)
        return cls._extra_fields

    model = None

    only_fields = None
    exclude_fields = None
    extra_fields = None
    field_permissions = None
    default_field_permissions = None

    default_page_size = None
    max_page_size = None

    list_action = ListAction()
    detail_action = DetailAction()

    create_action = CreateAction()
    update_action = UpdateAction()
    delete_action = DeleteAction()

    extra_actions = {}

    default_action_permissions = None
